from unittest.mock import sentinel, call, MagicMock

import pytest
from dopus_proxy import DOpusProxy, init_dop
import dopus_proxy

class TestDOpusProxy:
    def test_init(self):
        # given
        name = 'dummy'
        # when
        result = DOpusProxy(sentinel.obj, name, sentinel.parent)
        # then
        assert result._obj == sentinel.obj
        assert result._name == name
        assert result._parent == sentinel.parent

    def test_init_default(self):
        # when
        result = DOpusProxy(sentinel.obj)
        # then
        assert result._obj == sentinel.obj
        assert result._name == '<top parent object>'
        assert result._parent == sentinel.obj

    # @pytest.mark.xfail(raises=ValueError, strict=True)
    # def test_init_fail(self):
    #     # when
    #     DOpusProxy()

    def test_getattr(self, mocker):
        # given
        data = 'abc'
        obj_mock = MagicMock(test=data)
        dop = DOpusProxy(obj_mock)
        # when
        result = dop.test
        # then
        assert (result._obj, result._name, result._parent) == (data, 'test', obj_mock)

    def test_setitem(self):
        # given
        key = 'test'
        value = 'abc'
        obj_mock = MagicMock()
        dop = DOpusProxy(obj_mock)
        # when
        dop[key] = value
        # then
        assert obj_mock.mock_calls == [call.__setitem__(key, value)]

    def test_getitem_standard_syntax(self):
        # given
        key = 'test'
        obj_mock = MagicMock()
        obj_mock.__getitem__.return_value = sentinel.retval
        dop = DOpusProxy(obj_mock)
        # when
        result = dop[key]
        # then
        assert obj_mock.mock_calls == [call.__getitem__(key)]
        assert result == sentinel.retval

    def test_getitem_call_syntax(self):
        # given
        key = 'test'
        obj_mock = MagicMock()
        obj_mock.__getitem__.side_effect = TypeError
        obj_mock.return_value = sentinel.retval
        dop = DOpusProxy(obj_mock)
        # when
        result = dop[key]
        # then
        assert obj_mock.mock_calls == [call.__getitem__(key), call.__call__(key)]
        assert result == sentinel.retval

    def test_call_noargs(self):
        # given
        dop = DOpusProxy(sentinel.obj)
        # when
        result = dop()
        # then
        assert result._obj == sentinel.obj

    def test_call_args(self, mocker):
        # given
        parent_mock = MagicMock()
        dop = DOpusProxy(sentinel.obj, 'some_name', parent_mock)
        getattr_mock = mocker.patch('dopus_proxy.getattr')
        getattr_mock.return_value.return_value = sentinel.internal_call
        dop_mock = mocker.patch('dopus_proxy.DOpusProxy')
        dop_mock.return_value = sentinel.retval
        # when
        result = dop(1, arg='smth')
        # then
        assert parent_mock.method_calls == [call._FlagAsMethod('some_name')]
        assert getattr_mock.mock_calls == [call(parent_mock, 'some_name'), call()(1, arg='smth')]
        assert dop_mock.mock_calls == [call(sentinel.internal_call, '<call result>', sentinel.internal_call)]
        assert result == sentinel.retval

    def test_str(self):
        # given
        retval = "123test"
        obj_mock = MagicMock()
        obj_mock.__str__.return_value = retval
        dop = DOpusProxy(obj_mock)
        # when
        result = str(dop)
        # then
        assert result == retval

    def test_int(self):
        # given
        retval = 123
        obj_mock = MagicMock()
        obj_mock.__int__.return_value = retval
        dop = DOpusProxy(obj_mock)
        # when
        result = int(dop)
        # then
        assert result == retval

    def test_obj(self):
        # given
        dop = DOpusProxy(sentinel.obj)
        # when
        result = dop.obj
        # then
        assert result is sentinel.obj

    def test_setattr(self):
        # given
        obj_mock = MagicMock()
        dop = DOpusProxy(obj_mock)
        # when
        dop.test = 'abc'
        # then
        assert dop._obj.test == 'abc'


def test_init_dop(mocker):
    # given
    mocker.patch('dopus_proxy._do', new=sentinel.zero)
    # when
    init_dop(sentinel.smth)
    # then
    assert dopus_proxy._do == sentinel.smth


def test_to_vector(mocker):
    # given
    data = [1, 2]
    dop_mock = mocker.patch('dopus_proxy.DOpusProxy')
    # when
    dopus_proxy.to_vector(data)
    # then
    assert dop_mock.mock_calls == [call(None), call().Create(), call().Create().Vector(*data)]


def test_from_vector(mocker):
    # given
    data = [1, 2, 3]
    # when
    result = dopus_proxy.from_vector(data)
    # then
    assert result == data
    assert result is not data


def test_to_map(mocker):
    # given
    data = {1: 'ala', 'bela': 2}
    dop_mock = mocker.patch('dopus_proxy.DOpusProxy')
    # when
    dopus_proxy.to_map(data)
    # then
    assert dop_mock.mock_calls == [call(None),
                                   call().Create(),
                                   call().Create().Map(),
                                   call().Create().Map().__setitem__(1, 'ala'),
                                   call().Create().Map().__setitem__('bela', 2)]


def test_from_map():
    # given
    data = {1: 'ala', 'bela': 2}
    coll_mock = MagicMock(**{'__iter__.return_value': data.keys()}, side_effect=lambda x: data[x])
    # when
    result = dopus_proxy.from_map(coll_mock)
    # then
    assert result == data
    assert result is not data


def test_dout(mocker):
    # given
    data = 'abc'
    do_mock = mocker.patch('dopus_proxy._do')
    # when
    dopus_proxy.dout(data)
    # then
    assert do_mock.mock_calls == [call.Output(data)]